from __future__ import annotations

import enum
import io
import pathlib
import typing as tp
from string import ascii_letters

from athame import exceptions
from athame.trie import Trie
from pydantic import BaseModel


class TokenType(enum.Enum):
    END_INPUT = enum.auto()
    ALLOW = enum.auto()
    FORBID = enum.auto()
    DAY_OF_WEEK = enum.auto()
    TIME = enum.auto()
    DASH = enum.auto()


class Position(BaseModel):
    column: int
    lineno: int


class Token(BaseModel):
    type: TokenType
    lexeme: str
    position: Position


class Lexer:
    HOUR_STARTS = ("0", "1", "2")
    HOUR_ENDS = ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9")
    HOUR_ENDS_2X = ("0", "1", "2", "3")
    MINUTE_STARTS = ("0", "1", "2", "3", "4", "5")
    MINUTE_ENDS = ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9")
    DAYS_OF_WEEK = (
        "sunday",
        "sun",
        "monday",
        "mon",
        "tuesday",
        "tue",
        "wednesday",
        "wed",
        "thursday",
        "thu",
        "friday",
        "fri",
        "saturday",
        "sat",
    )
    KEYWORDS = Trie.from_strings("allow", "forbid", *DAYS_OF_WEEK)

    def __init__(self, source: tp.TextIO, *, whitespace: str = " \t\f", end_of_line: str = "\n"):
        self.source = source
        self.whitespace = set(whitespace)
        self.end_of_line = end_of_line
        self.line: str = self.source.readline()
        self.char: tp.Optional[str] = None
        self.lineno = 0
        self.column = -1
        self.token: tp.Optional[Token] = None
        self.position: Position = None

        self.next_char()

    @classmethod
    def from_file(cls, file: tp.Union[str, pathlib.Path], **kwargs) -> Lexer:
        if isinstance(file, str):
            file = pathlib.Path(file)

        return cls(file.resolve().open("r"), **kwargs)

    @classmethod
    def from_string(cls, string: str, **kwargs) -> Lexer:
        source = io.StringIO(string)
        return cls(source, **kwargs)

    def next_token(self) -> Token:
        self.token = None
        self.skip_whitespace()
        self.mark_position()

        if self.char is None:
            self.next_char()
            self.token = Token(type=TokenType.END_INPUT, lexeme="", position=self.position)
        elif Lexer.is_ascii_identifier_start(self.char):
            self.allow_or_forbid_or_day_of_week_token()
        elif Lexer.is_hour_start(self.char):
            self.time_token()
        elif self.char == "-":
            self.next_char()
            self.token = Token(type=TokenType.DASH, lexeme="-", position=self.position)

        if self.token is None:
            self.unexpected_char_error()

        return self.token

    def allow_or_forbid_or_day_of_week_token(self):
        lexeme = self.extract_ascii_identifier()

        if lexeme not in Lexer.KEYWORDS:
            if Lexer.KEYWORDS.contains_prefix(lexeme):
                expected = Lexer.KEYWORDS.chars_after(lexeme)
                self.unexpected_char_error(expected=expected)

            max_match = Lexer.KEYWORDS.max_match_for(lexeme)
            offset = len(max_match) - len(lexeme)
            self.unexpected_char_error(offset=offset)

        if lexeme.lower() == "allow":
            self.token = Token(type=TokenType.ALLOW, lexeme=lexeme, position=self.position)
        elif lexeme.lower() == "forbid":
            self.token = Token(type=TokenType.FORBID, lexeme=lexeme, position=self.position)
        elif lexeme.lower() in Lexer.DAYS_OF_WEEK:
            self.token = Token(type=TokenType.DAY_OF_WEEK, lexeme=lexeme, position=self.position)

    def time_token(self):
        hour_part = self.hour_partial()
        minute_part = self.minute_partial()
        lexeme = hour_part + minute_part
        self.token = Token(type=TokenType.TIME, lexeme=lexeme, position=self.position)

    def hour_partial(self) -> str:
        buffer = self.char
        self.next_char()
        if not Lexer.is_hour_end(self.char, hour=buffer):
            expected = Lexer.HOUR_ENDS_2X if buffer == "2" else Lexer.HOUR_ENDS
            self.unexpected_char_error(expected=expected)
        buffer += self.char
        self.next_char()
        return buffer

    def minute_partial(self) -> str:
        if not Lexer.is_minute_start(self.char):
            self.unexpected_char_error(expected=Lexer.MINUTE_STARTS)
        buffer = self.char
        self.next_char()
        if not Lexer.is_minute_end(self.char):
            self.unexpected_char_error(expected=Lexer.MINUTE_ENDS)
        buffer += self.char
        self.next_char()
        return buffer

    def skip_whitespace(self):
        while self.char in self.whitespace or self.char == self.end_of_line:
            if self.char == self.end_of_line:
                self.next_line()
            else:
                self.next_char()

    def next_line(self):
        self.line = self.source.readline()
        self.column = -1
        self.lineno += 1
        self.next_char()

    def next_char(self) -> str:
        if not self.line:
            self.column = 0
            self.char = None
        elif self.column + 1 >= len(self.line):
            self.char = self.end_of_line
            self.column += 1
        else:
            self.column += 1
            self.char = self.line[self.column]

        return self.char

    def extract_ascii_identifier(self) -> str:
        start = self.column
        while self.char and self.char in ascii_letters:
            self.next_char()
        return self.line[start : self.column]

    def mark_position(self):
        self.position = Position(column=self.column, lineno=self.lineno)

    def unexpected_char_error(self, *, expected: tp.Sequence[str] = None, offset: int = 0):
        column = self.column + offset
        message = f"missing or unexpected character in input stream on line {self.lineno} column {column}"
        message += "\n\n"
        message += self.build_error_string(column)

        if expected is not None:
            message += "\n\n"
            message += "expected one of: {}".format(", ".join(f"'{exp}'" for exp in expected))

        raise exceptions.LexerError(message)

    def build_error_string(self, column: int) -> str:
        # max line characters to display 80, prefer arrow left of center when truncating
        col_start = max(0, column - 39)
        line_part = self.line[col_start : col_start + 80].rstrip(self.end_of_line)
        arrow_part = " " * (column - col_start) + "^"
        padding = " " * 5
        error_string = padding + line_part
        error_string += "\n"
        error_string += padding + arrow_part
        return error_string

    def build_position_error_string(self) -> str:
        return self.build_error_string(self.position.column)

    @staticmethod
    def is_ascii_identifier_start(char: str) -> bool:
        return char in ascii_letters

    @staticmethod
    def is_hour_start(char: str) -> bool:
        return char in Lexer.HOUR_STARTS

    @staticmethod
    def is_hour_end(char: str, hour: str):
        return char in Lexer.HOUR_ENDS_2X if hour == "2" else char in Lexer.HOUR_ENDS

    @staticmethod
    def is_minute_start(char: str) -> bool:
        return char in Lexer.MINUTE_STARTS

    @staticmethod
    def is_minute_end(char: str) -> bool:
        return char in Lexer.MINUTE_ENDS



import pendulum
from athame.ast import Schedule
from athame.lexer import Lexer
from athame.parser import Parser


def from_dsl(string: str, tz: str = pendulum.local_timezone().name) -> Schedule:
    lexer = Lexer.from_string(string)
    parser = Parser(lexer)
    schedule = parser.parse()
    schedule.tz = tz
    return schedule



from __future__ import annotations

import typing as tp


class Trie:
    """
    A Trie (pronounced "try"), also known as a Prefix Tree, is a search tree for looking
    up specific keys from within a set. For our implementation our keys will be
    case-insensitive strings.

    More reading can be done here: https://en.wikipedia.org/wiki/Trie
    """

    def __init__(self):
        self.root = dict()

    @classmethod
    def from_strings(cls, *strings: tp.Sequence[str]) -> Trie:
        trie = cls()
        for string in strings:
            trie.add(string)
        return trie

    def __contains__(self, string: str) -> bool:
        string = string.lower()
        current = self.root
        for char in string:
            if (current := current.get(char)) is None:
                return False
        return current.get("is_terminal")

    def add(self, string: str):
        string = string.lower()
        current = self.root
        for char in string:
            current = current.setdefault(char, dict(is_terminal=False))
        current["is_terminal"] = True

    def contains_prefix(self, string: str) -> bool:
        string = string.lower()
        current = self.root
        for char in string:
            if (current := current.get(char)) is None:
                return False
        return True

    def chars_after(self, string: str) -> tp.List[str]:
        string = string.lower()
        current = self.root
        for char in string:
            if (current := current.get(char)) is None:
                return []
        return sorted(char for char in current if char != "is_terminal")

    def max_match_for(self, string: str) -> str:
        string = string.lower()
        buffer = ""
        current = self.root
        for char in string:
            if (current := current.get(char)) is None:
                return buffer
            buffer += char
        return buffer



import typing as tp

from athame import exceptions
from athame.ast import Agenda, Directive, Interval, Schedule, Time
from athame.lexer import Lexer, TokenType


class Parser:
    GRAMMAR = """
    schedule  : schedule days agendas
              | days agendas

    days      : DAY_OF_WEEK '-' DAY_OF_WEEK
              | DAY_OF_WEEK

    agendas   : agendas agenda
              | agenda

    agenda    : ALLOW intervals
              | FORBID intervals

    intervals : intervals interval
              | interval

    interval  : TIME '-' TIME
              | '-' TIME
              | TIME
    """

    def __init__(self, lexer: Lexer):
        self.lexer = lexer

    def parse(self) -> Schedule:
        self.lexer.next_token()
        return self.parse_schedule()

    def parse_schedule(self) -> Schedule:
        """
        schedule : schedule days agendas
                 | days agendas
        """

        schedule = Schedule()
        days = self.parse_days()
        agendas = self.parse_agendas()
        schedule.add_agendas_on_days(agendas, days)
        while self.lexer.token.type is TokenType.DAY_OF_WEEK:
            days = self.parse_days()
            agendas = self.parse_agendas()
            schedule.add_agendas_on_days(agendas, days)

        return schedule

    def parse_days(self) -> tp.List[int]:
        """
        days : DAY_OF_WEEK '-' DAY_OF_WEEK
             | DAY_OF_WEEK
        """

        if self.lexer.token.type is not TokenType.DAY_OF_WEEK:
            self.unexpected_token(expected=[TokenType.DAY_OF_WEEK])

        dow_int_map = dict(
            sunday=0,
            sun=0,
            monday=1,
            mon=1,
            tuesday=2,
            tue=2,
            wednesday=3,
            wed=3,
            thursday=4,
            thu=4,
            friday=5,
            fri=5,
            saturday=6,
            sat=6,
        )
        days = [dow_int_map.get(self.lexer.token.lexeme.lower())]
        self.lexer.next_token()
        if self.lexer.token.type is TokenType.DASH:
            self.lexer.next_token()
            if self.lexer.token.type is not TokenType.DAY_OF_WEEK:
                self.unexpected_token(expected=[TokenType.DAY_OF_WEEK])
            last_day = dow_int_map.get(self.lexer.token.lexeme.lower())
            self.lexer.next_token()
            next_day = (days[0] + 1) % 7
            while days[-1] != last_day:
                days.append(next_day)
                next_day = (next_day + 1) % 7

        return days

    def parse_agendas(self) -> tp.List[Agenda]:
        """
        agendas : agendas agenda
                | agenda
        """

        agendas = []
        agendas.append(self.parse_agenda())
        while self.lexer.token.type in {TokenType.ALLOW, TokenType.FORBID}:
            agendas.append(self.parse_agenda())

        return agendas

    def parse_agenda(self) -> Agenda:
        """
        agenda : ALLOW intervals
               | FORBID intervals
        """

        if self.lexer.token.type is TokenType.ALLOW:
            self.lexer.next_token()
            intervals = self.parse_intervals()
            agenda = Agenda(intervals=intervals, directive=Directive.ALLOW)
        elif self.lexer.token.type is TokenType.FORBID:
            self.lexer.next_token()
            intervals = self.parse_intervals()
            agenda = Agenda(intervals=intervals, directive=Directive.FORBID)
        else:
            self.unexpected_token(expected=[TokenType.ALLOW, TokenType.FORBID])

        return agenda

    def parse_intervals(self) -> tp.List[Interval]:
        """
        intervals : intervals interval
                  | interval
        """

        intervals = []
        intervals.append(self.parse_interval())
        while self.lexer.token.type is TokenType.TIME or self.lexer.token.type is TokenType.DASH:
            intervals.append(self.parse_interval())

        return intervals

    def parse_interval(self) -> Interval:
        """
        interval : TIME '-' TIME
                 | '-' TIME
                 | TIME
        """

        if self.lexer.token.type is TokenType.TIME:
            start = Time.from_string(self.lexer.token.lexeme)
            self.lexer.next_token()
            if self.lexer.token.type is TokenType.DASH:
                self.lexer.next_token()
                if self.lexer.token.type is not TokenType.TIME:
                    self.unexpected_token(expected=[TokenType.TIME])
                end = Time.from_string(self.lexer.token.lexeme)
                self.lexer.next_token()
            else:
                end = Time.end_of_day()

        elif self.lexer.token.type is TokenType.DASH:
            self.lexer.next_token()
            if self.lexer.token.type is not TokenType.TIME:
                self.unexpected_token(expected=[TokenType.TIME])
            start = Time.start_of_day()
            end = Time.from_string(self.lexer.token.lexeme)
            self.lexer.next_token()

        else:
            self.unexpected_token(expected=[TokenType.TIME, TokenType.DASH])

        return Interval(start=start, end=end)

    def unexpected_token(self, expected: tp.Sequence[TokenType]):
        message = f"missing or unexpected token in token stream on line {self.lexer.lineno} column {self.lexer.column}"
        message += "\n\n"
        message += self.lexer.build_position_error_string()
        message += "\n\n"
        message += "expected one of: {}".format(", ".join(token_type.name for token_type in expected))
        message += "\n"
        message += f"got: {self.lexer.token.type.name}"
        raise exceptions.ParserError(message)



from __future__ import annotations

import enum
import itertools
import typing as tp
from time import sleep

import pendulum
from athame import exceptions
from pydantic import BaseModel, Field, conint


class Time(BaseModel):
    hour: conint(ge=0, le=23)
    minute: conint(ge=0, le=59)

    @classmethod
    def from_string(cls, string: str) -> Time:
        return Time(hour=int(string[:2]), minute=int(string[2:]))

    @classmethod
    def start_of_day(cls) -> Time:
        return cls(hour=0, minute=0)

    @classmethod
    def end_of_day(cls) -> Time:
        return cls(hour=23, minute=59)


class Interval(BaseModel):
    start: Time
    end: Time


# normally I would define this as a member of Agenda but
# 'from __future__ import annotations' forces us to do it this way
class Directive(enum.Enum):
    ALLOW = enum.auto()
    FORBID = enum.auto()


class Agenda(BaseModel):
    intervals: tp.List[Interval]
    directive: Directive


class WeekdayInterval(BaseModel):
    """
    Interval of time within a single day of the week.
    """

    day_of_week: conint(ge=0, le=6)
    interval: Interval

    def __contains__(self, dt: pendulum.DateTime):
        """
        Determines if a datetime is within this weekday interval.

        The boolean logic is as follows:
            - First make sure the day of the week is correct.
            - Then check that one of the following conditions is true:
                1. The datetime hour is between (exclusive) the start and end hour.
                2. The datetime hour is equal to the starting hour, not equal to the ending hour,
                   and the datetime minute is at or after the starting minute.
                3. The datetime hour is equal to the ending hour, not equal to the starting hour,
                   and the datetime minute is at or before the ending minute.
                4. The datetime hour is equal to both the starting and ending hour and the
                   datetime minute is between (inclusive) the starting and ending minutes.
        """
        return self.day_of_week == dt.day_of_week and (
            self.interval.start.hour < dt.hour < self.interval.end.hour
            or self.interval.start.hour == dt.hour != self.interval.end.hour
            and self.interval.start.minute <= dt.minute
            or self.interval.start.hour != dt.hour == self.interval.end.hour
            and self.interval.end.minute >= dt.minute
            or self.interval.start.hour == self.interval.end.hour == dt.hour
            and self.interval.start.minute <= dt.minute <= self.interval.end.minute
        )

    def next_start_after(self, moment: pendulum.DateTime) -> pendulum.DateTime:
        """
        Returns the next datetime at the start of this interval after the given moment.

        Example:
            In this example we have an interval which starts at 00:00 on Mondays. When we ask for its next start after
            00:00 on 2017-01-01 (a Sunday) it returns 00:00 on 2017-01-02.

            >>> interval = Interval(start=Time(hour=0, minute=0), end=Time.end_of_day())
            >>> weekday_interval = WeekdayInterval(day_of_week=1, interval=interval)
            >>> moment = pendulum.datetime(2017, 1, 1, hour=0, minute=0)  # sunday
            >>> weekday_interval.next_start_after(moment)
            DateTime(2017, 1, 2, 0, 0, ...)

            If we were to ask it for the next start after a time on a Tuesday (or a time after 00:00 on a Monday) it would
            return 00:00 the following Monday.
        """
        # since pendulum.DateTime.next never returns the current day, we must manually check if the next occurence
        # of <target> is today when the days of week match.
        if moment.day_of_week == self.day_of_week:
            time_shifted_moment = moment.at(hour=self.interval.start.hour, minute=self.interval.start.minute)
            return (
                time_shifted_moment
                if time_shifted_moment >= moment
                else time_shifted_moment.next(self.day_of_week, keep_time=True)
            )

        return moment.next(self.day_of_week).at(hour=self.interval.start.hour, minute=self.interval.start.minute)

    def next_end_after(self, moment: pendulum.DateTime) -> pendulum.DateTime:
        """
        Returns the next datetime *after* the next end of this interval after the given moment. That is to say, 1 minute after
        the next end of this interval.

        Example:
            In this example we have an interval which ends at 01:30 on Mondays. When we ask for its next end after
            00:00 on 2017-01-01 (a Sunday) it returns 01:31 on 2017-01-02.

            >>> interval = Interval(start=Time.start_of_day(), end=Time(hour=1, minute=30))
            >>> weekday_interval = WeekdayInterval(day_of_week=1, interval=interval)
            >>> moment = pendulum.datetime(2017, 1, 1, hour=0, minute=0)  # sunday
            >>> weekday_interval.next_end_after(moment)
            DateTime(2017, 1, 2, 1, 31, ...)

            If we were to ask it for the next end after a time on a Tuesday (or a time after 01:31 on a Monday) it would
            return 01:31 the following Monday.
        """
        # since pendulum.DateTime.next never returns the current day, we must manually check if the next occurence
        # of <target> is today when the days of week match.
        if moment.day_of_week == self.day_of_week:
            time_shifted_moment = moment.at(hour=self.interval.end.hour, minute=self.interval.end.minute).add(minutes=1)
            return (
                time_shifted_moment
                if time_shifted_moment > moment
                else time_shifted_moment.next(self.day_of_week, keep_time=True)
            )

        return (
            moment.next(self.day_of_week)
            .at(hour=self.interval.end.hour, minute=self.interval.end.minute)
            .add(minutes=1)
        )


class Schedule(BaseModel):
    """
    This is a class for specifying intervals of time during the week in which a background worker
    is allowed to perform its work. This differs from a tool like cron in that cron is designed to
    unconditionally kick off the execution of a task at certain times whereas this is to designed
    specify the intervals of time during a week when a endlessly running worker is permitted to
    execute.
    """

    allowed: tp.List[WeekdayInterval] = Field(default_factory=list)
    forbidden: tp.List[WeekdayInterval] = Field(default_factory=list)
    tz: str = pendulum.local_timezone().name

    @classmethod
    def always_allowed(cls, tz: str = pendulum.local_timezone().name) -> Schedule:
        allowed = [
            WeekdayInterval(
                day_of_week=day_of_week,
                interval=Interval(
                    start=Time.start_of_day(),
                    end=Time.end_of_day(),
                ),
            )
            for day_of_week in range(7)
        ]
        tz = tz or pendulum.local_timezone().name
        return cls(allowed=allowed, tz=tz)

    def add_agenda_on_days(self, agenda: Agenda, days: tp.List[int]):
        weekday_intervals = [
            WeekdayInterval(day_of_week=day_of_week, interval=interval)
            for interval, day_of_week in itertools.product(agenda.intervals, days)
        ]
        if agenda.directive is Directive.ALLOW:
            self.allowed.extend(weekday_intervals)
        elif agenda.directive is Directive.FORBID:
            self.forbidden.extend(weekday_intervals)
        else:
            raise ValueError("invalid agenda type")

    def add_agendas_on_days(self, agendas: tp.List[Agenda], days: tp.List[int]):
        for agenda in agendas:
            self.add_agenda_on_days(agenda, days)

    def is_allowed_at(self, moment: pendulum.DateTime) -> bool:
        moment = moment.in_tz(self.tz)
        return any(moment in weekday_interval for weekday_interval in self.allowed) and not any(
            moment in weekday_interval for weekday_interval in self.forbidden
        )

    def is_allowed_at_or_raise(self, moment: pendulum.DateTime):
        if not self.is_allowed_at(moment):
            raise exceptions.ScheduleError(f"{moment.format('dddd HH:mm')} is not allowed")

    def is_allowed_now(self) -> bool:
        return self.is_allowed_at(pendulum.now())

    def is_allowed_now_or_raise(self):
        self.is_allowed_at_or_raise(pendulum.now())

    def next_moment_allowed_after(self, moment: pendulum.DateTime) -> pendulum.DateTime:
        # the next allowed moment will always be among the following candidates:
        #   1. the given moment
        #   2. the start of an allowed interval
        #   3. the end of a forbidden interval
        moment = moment.in_tz(self.tz)
        allowed_starts = [wdi.next_start_after(moment) for wdi in self.allowed]
        forbidden_ends = [wdi.next_end_after(moment) for wdi in self.forbidden]
        candidates = [moment] + allowed_starts + forbidden_ends
        allowed_only = filter(self.is_allowed_at, candidates)
        with exceptions.ScheduleError.handle_errors("no allowed intervals in the schedule", exception_class=ValueError):
            best, *_ = sorted(allowed_only)

        return best

    def sleep_until_allowed(self):
        moment = pendulum.now()

        if self.is_allowed_at(moment):
            return

        next_allowed_moment = self.next_moment_allowed_after(moment)
        sleep_duration = next_allowed_moment - moment
        sleep(sleep_duration.total_seconds())
